import numpy as np
import sasktran as sk
from sasktran_tir.constants import Boltzmann
from sasktran_tir.util import _atm_reader, _atm_file_path


class ClimatologyAtmosphericState(sk.ClimatologyUserDefined):
    """
    Creates a custom climatology, containing temperature and pressure, for use as an atmospheric state.
    The profiles are from http://eodg.atm.ox.ac.uk/RFM/atm/ and are divided into three folders:
    'FASCODE/ICRCCM Model Atmospheres', 'MIPAS Model Atmospheres (1998)', and 'MIPAS Model Atmospheres (2001)'.
    Each folder contains several '.atm' files, where each file contains a set of profiles describing various
    latitudinal regions and times.

    NOTE: These are one-dimensional atmospheric profiles. Latitude, longitude, and mjd will not affect the returned values.
    """

    def __init__(self, dataset: str = 'fascode', climatology: str = 'std', interp: str = 'linear',
                 spline: bool = False):
        """
        Create an atmospheric state from data files. By default uses the US Standard Atmosphere.

        Parameters
        ----------
        dataset : str, optional
            Name of dataset to use. Valid choices are 'fascode', 'mipas_1998', and 'mipas_2001'. Default is 'fascode'.
        climatology : str, optional
            Name of the data file within the chosen dataset to load. Default is 'std'.
        interp : str, optional
            One of 'linear' or 'log'. Defines the interpolation space. Default is 'linear'.
        spline: bool, optional
            One of True or False.  If True then a quadratic spline will be used to interpolate, if False the piecewise
            linear interpolation is done.  Default is False.

        Examples
        --------
        >>> from sasktran_tir.climatology import ClimatologyAtmosphericState
        >>> us_standard_atmosphere = ClimatologyAtmosphericState()  # default atmosphere
        >>> std_temperature = us_standard_atmosphere.get_parameter('SKCLIMATOLOGY_TEMPERATURE_K', latitude=0, longitude=0, altitudes=np.linspace(20000, 30000, 4), mjd=54372)
        >>> print(std_temperature)
        [216.7        219.93333333 223.2        226.5       ]
        >>> mipas_2001_polar_summer = ClimatologyAtmosphericState(atm_folder='mipas_2001', atm_file='sum')
        >>> polar_pressure = mipas_2001_polar_summer.get_parameter('SKCLIMATOLOGY_PRESSURE_PA', latitude=0, longitude=0, altitudes=np.linspace(20000, 30000, 4), mjd=54372)
        >>> print(polar_pressure)
        [5537.54       3435.81       2128.99333333 1329.72      ]
        """

        # get climatology data
        if climatology.endswith('.atm'):
            atm_file = climatology
        else:
            atm_file = climatology + '.atm'
        file_path = _atm_file_path(dataset, atm_file)
        data = _atm_reader(file_path)

        super().__init__(data['HGT'] * 1000.0,
                         {'SKCLIMATOLOGY_TEMPERATURE_K': data['TEM'], 'SKCLIMATOLOGY_PRESSURE_PA': data['PRE'] * 100.0},
                         interp=interp, spline=spline)


class ClimatologySpecies(sk.ClimatologyUserDefined):
    """
    Create a custom climatology using a species profile from data files. By default uses the US Standard Atmosphere.
    The CLIMATOLOGY_HANDLE of the species will be SKCLIMATOLOGY_<species>_CM3. Note that despite the files
    providing profiles of species in PPMV (parts per million by volume) the climatology created will be a profile
    with units of molecules per cm^3. This is because the VMR climatologies are currently not supported in the
    SASKTRAN cross section calculation. The species PPMV is converted to using the temperature, :math:`T`, and
    pressure, :math:`p`, in the specified '.atm' file and using the Ideal Gas Law to compute a neutral density,
    :math:`n`:

    .. math::

        n = p / (k T)

    where k is the Boltzmann constant.  The species VMR is multiplied by the neutral density to obtain the species
    density.

    NOTE: These are one-dimensional atmospheric profiles. Latitude, longitude, and MJD will not affect the returned
    values.

    Parameters
    ----------
    species : str
        Chemical formula of the species to use. E.g. 'O3', 'CO2', etc.
    dataset : str, optional
        Name of dataset to use. Valid choices are 'fascode', 'mipas_1998', and 'mipas_2001'. Default is 'fascode'.
    climatology : str, optional
        Name of the data file within the chosen dataset to load. Default is 'std'.
    interp : str, optional
        One of 'linear' or 'log'. Defines the interpolation space. Default is 'linear'.
    spline: bool, optional
        One of True or False.  If True then a quadratic spline will be used to interpolate, if False the piecewise
        linear interpolation is done.  Default is False.

    Examples
    --------
    >>> import numpy as np
    >>> import sasktran as sk
    >>> from sasktran_tir.climatology import ClimatologySpecies
    >>> mipas_2001_polar_summer_co2 = ClimatologySpecies('CO2', atm_folder='mipas_2001', atm_file='sum')
    >>> n_co2 = mipas_2001_polar_summer_co2.get_parameter('SKCLIMATOLOGY_CO2_CM3', latitude=0, longitude=0, altitudes=np.arange(0, 100000, 500), mjd=54372)
    >>> print(n_co2)
    [6.17180616e+14 3.82011862e+14 2.35914675e+14 1.41142413e+14]


    Species Supported by Each Climatology
    -------------------------------------

    Each dataset contains several different climatologies. Each climatology contains unique profiles of temperature,
    pressure, and VMRs of the major species. The VMRs for the minor species are the same for all climatologies in each
    dataset. However, the number density of each species is computed using the pressure and temperature of the selected
    climatology.

    +------------+---------------------------------+---------------------------------+---------------------------------+
    | Dataset    | Climatologies                   | Major Species                   | Minor Species                   |
    +============+=================================+=================================+=================================+
    | fascode    | tro (Tropical),                 | H2O, CO2, O3, N2O, CO, CH4, O2  | NO, SO2, NO2, NH3, HNO3, OH,    |
    |            | mls (Mid-Latitude Summer),      |                                 | HF, HCl, HBr, HI, ClO, OCS,     |
    |            | mlw (Mid-Latitude Winter),      |                                 | H2CO, HOCl, N2, HCN, CH3Cl,     |
    |            | sas (Sub-Arctic Summer),        |                                 | H2O2, C2H2, C2H6, PH3, COF2,    |
    |            | saw (Sub-Arctic Winter),        |                                 | SF6, H2S, CFCl3, CF2Cl2, CClF3, |
    |            | std (US Standard Atmosphere),   |                                 | CF4, CHCl2F, CHClF2, C2Cl3F3,   |
    |            |                                 |                                 | C2Cl2F4, C2ClF5, CCl4, ClONO2,  |
    |            |                                 |                                 | N2O5, HNO4                      |
    +------------+---------------------------------+---------------------------------+---------------------------------+
    | mipas_1998 | day_imk (Mid-Latitude Day)      | N2, O2, O3P, CO2, O3, H2O, CH4, | HCN, H2O2, F12, F14, F22, COF2, |
    |            | ngt_imk (Mid-Latitude Night)    | N2O, HNO3, CO, NO2, N2O5, ClO,  | OCS, NH3, SO2, CFCl3, C2H2,     |
    |            | win_imk (Polar Winter)          | HOCl, ClONO2, NO                | C2H6, CCl4, SF6, HNO4, CH3Cl,   |
    |            | sum_imk (Polar Summer)          |                                 | CClF3, CHCl2F, C2Cl3F3, C2Cl2F4 |
    +------------+---------------------------------+---------------------------------+---------------------------------+
    | mipas_2001 | day (Mid-Latitude Day)          | N2, O2, CO2, O3, H2O, CH4, N2O, | CClF3, CHCl2F, C2Cl3F3,         |
    |            | ngt (Mid-Latitude Night)        | HNO3, CO, NO2, N2O5, ClO, HOCl, | C2Cl2F4, C2ClF5, CH3Cl, H2S     |
    |            | win (Polar Winter)              | ClONO2, NO, HNO4, HCN, NH3,     |                                 |
    |            | sum (Polar Summer)              | F11, F12, F14, F22, CCl4, COF2, |                                 |
    |            | equ (Equatorial Day)            | H2O2, C2H2, C2H6, OCS, SO2, SF6 |                                 |
    +------------+---------------------------------+---------------------------------+---------------------------------+
    """

    def __init__(self, species: str, dataset: str = 'fascode', climatology: str = 'std', interp: str = 'linear',
                 spline: bool = False):
        # get climatology data
        if climatology.endswith('.atm'):
            atm_file = climatology
        else:
            atm_file = climatology + '.atm'
        file_path = _atm_file_path(dataset, atm_file)
        data = _atm_reader(file_path)

        if species in data:
            # Calculate neutral density from Ideal Gas Law, n = p / (k T) then convert from m^-3 to cm^-3
            air_nd = data['PRE'] * 100.0 / Boltzmann / data['TEM'] / 1000000.0
            species_nd = data[species] / 1.0e6 * air_nd
        else:
            # This is a minor species, need to load the minor species data
            if dataset == 'mipas_2001':
                file_path = _atm_file_path(dataset, 'extra.atm')
            elif dataset == 'mipas_1998':
                file_path = _atm_file_path(dataset, 'extra_imk.atm')
            elif dataset == 'fascode':
                file_path = _atm_file_path(dataset, 'minor.atm')
            else:
                raise ValueError('{} is not a valid dataset'.format(dataset))
            data_minor = _atm_reader(file_path)
            pre_minor = np.interp(data_minor['HGT'], data['HGT'], data['PRE'])  # get pressure at minor species heights
            tem_minor = np.interp(data_minor['HGT'], data['HGT'], data['TEM'])
            air_nd_minor = pre_minor * 100.0 / Boltzmann / tem_minor / 1000000.0
            species_nd = data_minor[species] / 1.0e6 * air_nd_minor

        super().__init__(data['HGT'] * 1000.0,
                         {'SKCLIMATOLOGY_' + species + '_CM3': species_nd}, interp=interp, spline=spline)


class ClimatologyFull(sk.ClimatologyUserDefined):
    """
    Creates a custom climatology from data files. Similar to ``ClimatologySpecies`` and ``ClimatologyAtmosphericState``
    but includes ALL profiles defined in the data file. See ``ClimatologySpecies`` for the species contained in each
    file.

    Parameters
    ----------
    dataset : str, optional
        Name of dataset to use. Valid choices are 'fascode', 'mipas_1998', and 'mipas_2001'. Default is 'fascode'.
    climatology : str, optional
        Name of the data file within the chosen dataset to load. Default is 'std'.
    include_minor : bool, optional
        If True, minor species are included in the climatology. Default is True.
    interp : str, optional
        One of 'linear' or 'log'. Defines the interpolation space. Default is 'linear'.
    spline: bool, optional
        One of True or False.  If True then a quadratic spline will be used to interpolate, if False the piecewise
        linear interpolation is done.  Default is False.
    """

    def __init__(self, dataset: str = 'fascode', climatology: str = 'std', include_minor=True, interp: str = 'linear',
                 spline: bool = False):

        # get climatology data
        if climatology.endswith('.atm'):
            atm_file = climatology
        else:
            atm_file = climatology + '.atm'
        file_path = _atm_file_path(dataset, atm_file)
        data = _atm_reader(file_path)

        # Calculate neutral density from Ideal Gas Law, n = p / (k T) then convert from m^-3 to cm^-3
        air_nd = data['PRE'] * 100.0 / Boltzmann / data['TEM'] / 1000000.0

        values = dict()
        for species in data.keys():
            if species == 'HGT':
                continue
            elif species == 'TEM':
                values['SKCLIMATOLOGY_TEMPERATURE_K'] = data['TEM']
            elif species == 'PRE':
                values['SKCLIMATOLOGY_PRESSURE_PA'] = data['PRE'] * 100.0
            else:
                values['SKCLIMATOLOGY_' + species + '_CM3'] = data[species] / 1.0e6 * air_nd

        if include_minor:
            if dataset == 'mipas_2001':
                file_path = _atm_file_path(dataset, 'extra.atm')
            elif dataset == 'mipas_1998':
                file_path = _atm_file_path(dataset, 'extra_imk.atm')
            elif dataset == 'fascode':
                file_path = _atm_file_path(dataset, 'minor.atm')
            else:
                raise ValueError('{} is not a valid dataset'.format(dataset))
            data_minor = _atm_reader(file_path)
            pre_minor = np.interp(data_minor['HGT'], data['HGT'], data['PRE'])  # get pressure at minor species heights
            tem_minor = np.interp(data_minor['HGT'], data['HGT'], data['TEM'])
            air_nd_minor = pre_minor * 100.0 / Boltzmann / tem_minor / 1000000.0

            for species in data_minor.keys():
                if species == 'HGT':
                    continue
                else:
                    values['SKCLIMATOLOGY_' + species + '_CM3'] = data_minor[species] / 1.0e6 * air_nd_minor

        super().__init__(data['HGT'] * 1000.0, values, interp=interp, spline=spline)
