"""
Tests that sktran_do modules is operating as expected.

Date: March 28, 2017
Geometry: TEMPO at (20, -100, 35786000), looking at latitudes [-38.51, 26.05], longitudes [-100, -100]
"""

import unittest, math
import numpy as np
import sasktranif.sasktranif as skif

import sys
def printf(format, *args):
    sys.stdout.write(format % args)

class QuickTest(unittest.TestCase):
    """
    Test bench to run sanity checks on SASKTRAN-DISCO. Note that the primary function of these tests if verifying the 
    SKIF interface as well as the internal interface between the SASKTRAN framework and the DISCO kernel (eg. making 
    sure species are being parameterized properly for the DISCO kernel). This test bench also is used to track 
    continuity between releases (eg. track that radiances between releases are the same).
    
    The primary test bench for DISCO is the ``sasktran_disco_engine_tests`` which is a program that runs lots of edge-
    case tests, but only verifies that the DISCO kernel is operating properly.
    """

    def test_weighting_functions(self):
        """ Compare SKDO weighting functions to those calculate with the finite-difference method.
        """
        engine = skif.ISKEngine('DO')
        engine.SetProperty('numstreams', 16)
        engine.SetProperty('numlayers', 25)
        engine.SetProperty('losdiagnostics', 1)

        # First create the simple atmosphere
        msis90 = skif.ISKClimatology('MSIS90')
        labow_ozone = skif.ISKClimatology('O3LABOW')
        opt_rayleigh = skif.ISKOpticalProperty('RAYLEIGH')
        opt_ozone = skif.ISKOpticalProperty('O3_OSIRISRES')
        wavelengths = [310]

        engine.SetAtmosphericState(msis90)
        engine.AddSpecies( 'SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3', msis90, opt_rayleigh)
        engine.AddSpecies( 'SKCLIMATOLOGY_O3_CM3', labow_ozone, opt_ozone)
        engine.SetWavelengths(wavelengths)

        # Configure surface
        lambertian_brdf = skif.ISKBrdf('Lambertian')
        lambertian_brdf.SetProperty('Albedo', 0.3)

        engine.SetBRDF(lambertian_brdf)

        # Configure geometry
        sun = [-0.190809, -0.98162718, 0]
        obs = [-6880579.70829706, -39021706.60809562, 14407230.65919554]
        look = [0.15042014, 0.87062868, -0.46837964]
        mjd = 54372.5

        engine.SetProperty('sun', sun)
        engine.AddLineOfSight(mjd, obs, look)

        # Configure weighting functions specifications:
        wf_altitudes = np.linspace(1e3, 51e3, 26)           # altitudes to calculate weighting functions at
        wf_widths = np.array([1000] * len(wf_altitudes))    # width of perturbation (above and below the given altitude)
        wf_eps = np.array([0.01] * len(wf_altitudes))       # relative change in species number density (perturbation)

        engine.SetProperty('wfclimatologies', 'SKCLIMATOLOGY_O3_CM3')
        engine.SetProperty('wfaltitudes', wf_altitudes)
        engine.SetProperty('wfwidths', wf_widths)
        engine.SetProperty('wfeps', wf_eps)
        engine.SetProperty('wfform', 0)                 # we want dI_dLogX


        engine.InitializeModel()
        ok, rad = engine.CalculateRadiance()
        ok, wf = engine.GetWeightingFunctions()

        wf_correct = np.array([
            # May 19, 2017
            -8.16129033e-05, -1.22650298e-04, -1.48115362e-04, -1.60420953e-04,
            -1.80710717e-04, -2.31164826e-04, -2.85136908e-04, -4.04224888e-04,
            -6.72436706e-04, -1.06773579e-03, -1.39997215e-03, -1.63143755e-03,
            -1.81549434e-03, -1.78586686e-03, -1.57352013e-03, -1.26769933e-03,
            -1.00312499e-03, -7.55823885e-04, -5.84603170e-04, -4.13461450e-04,
            -2.74406814e-04, -1.72550733e-04, -1.06845920e-04, -6.74153699e-05,
            -4.31475030e-05, -2.77138122e-05
        ])
        self.assertEqual(len(wf.flatten()), len(wf_correct))
        for w, c in zip(wf.flatten(), wf_correct):
            self.assertAlmostEqual(w, c)


    def test_diagnostics(self):
        """ Check that SKDO diagnostics are working, and internal values are correct.
        """
        geo = skif.GEODETIC_INSTANT()
        geo.FromSequence([0, 0, 0, 54372])

        clim_air = skif.ISKClimatology('MSIS90')
        opt_air = skif.ISKOpticalProperty('RAYLEIGH')

        engine = skif.ISKEngine("DO")
        
        engine.SetProperty('numstreams', 40)
        engine.SetProperty('numlayers', 200)

        engine.AddSpecies('SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3', clim_air, opt_air)
        engine.AddSpecies('SKCLIMATOLOGY_O3_CM3', skif.ISKClimatology("O3LABOW"),
                          skif.ISKOpticalProperty("O3_OSIRISRES"))

        engine.SetAtmosphericState(clim_air)

        sun = np.array([-0.190809, -0.98162718, 0])
        engine.SetProperty('sun', sun)

        obs = np.array([ -6880579.70829706, -39021706.60809562, 14407230.65919554])
        looks = np.array([[ 0.15042014, 0.87062868, -0.46837964],
                          [ 0.1518094, 0.87854755, -0.45288863]])
        mjd = 54372.5
        for los in np.reshape(looks, (-1, 3)):
            engine.AddLineOfSight(mjd, obs, los)

        engine.SetWavelengths([350, 400, 700])

        lambertian_brdf = skif.ISKBrdf('Lambertian')
        lambertian_brdf.SetProperty('Albedo', 0.3)

        engine.SetBRDF(lambertian_brdf)

        engine.SetProperty('diagnostics', 1.0)

        engine.InitializeModel()
        ok, rad = engine.CalculateRadiance()

        # Reconstruct the solutions
        for w in [0, 1, 2]: # wavelengths: 350, 400, 700
            for l in [0, 1]: # 2 look directions
                ok = True
                # Geometry
                ok, ref = engine.GetProperty('referencepoint[%i]' % l)
                ok, solar_zenith =  engine.GetProperty('localsolarzenith[%i]' % l)
                ok, los_zenith = engine.GetProperty('localloszenith[%i]' % l)
                ok, rel_azimuth = engine.GetProperty('localrelativeazimuth[%i]' % l)
                # Atmospheric properties
                num_los = 2
                wl_index = w * num_los + l
                ok, optical_depths = engine.GetProperty('daopticaldepths[%i, %i]' % (w, l))
                ok, single_scatter_albedo = engine.GetProperty('dassa[%i, %i]' % (w, l))
                ok, phasef_expansion = engine.GetProperty('daphasef[%i, %i]' % (w, l))
                # Radiative transfer solutions
                ok, radiance_components = engine.GetProperty('radiancecomponents[%i, %i]' % (w, l))
                ok, reflected_components = engine.GetProperty('reflectedcomponents[%i, %i]' % (w, l))
                ok, participating_source_terms = engine.GetProperty('participatingsourceterms[%i, %i]' % (w, l))
                # Calculate radiance
                radiance = 0
                for m in range(0, 40):
                    component = reflected_components[m] # radiance at the ground

                    for od, pst in zip(optical_depths[::-1], participating_source_terms[-40+m::-40]): # integrate up to TOA
                        component = component * math.exp(-od/math.cos(los_zenith)) + pst
                    component *= math.cos(m * rel_azimuth)
                    self.assertAlmostEqual(component, radiance_components[m])
                    radiance += component
                self.assertAlmostEqual(radiance, rad[w][l])

    def test_averaged_reference_point(self):
        """ Check the "averagereferencepoint" option is working correctly.
        For all the lines of sight to be solved with a single R.T. solution, a non-trivial sorting
        takes place. This test checks that nothing gets mixed up from that sorting.
        """
        try:
            import sasktran as sk
            import time
        except:
            import logging
            msg = "'averagereferencepoint' option will not be tested because the 'sasktran' library is not installed. To install it" \
                  " enter: 'pip install sasktran -f https://arg.usask.ca/wheels/"

            logging.getLogger().log(logging.WARNING, msg)
            return
        
        geometry = sk.NadirGeometry()
        zeniths = np.array([0, 70, 10, 60, 20, 50, 30, 40]).astype(float) # order should be neither increasing, nor decreasing
        azimuths = np.arange(0, 181, 10).astype(float)
        zeniths, azimuths = np.meshgrid(zeniths, azimuths, indexing='ij')
        geometry.from_zeniths_and_azimuth_difference(55, zeniths.flatten(), azimuths.flatten())
        geometry.reference_point = None

        atmosphere = sk.Atmosphere()
        atmosphere['rayleigh'] = sk.Species(sk.Rayleigh(), sk.MSIS90())
        atmosphere['o3'] = sk.Species(sk.O3OSIRISRes(), sk.Labow())
        atmosphere['no2'] = sk.Species(sk.NO2OSIRISRes(), sk.Pratmo())
        atmosphere.wf_species = 'no2'
        atmosphere.brdf = sk.Kokhanovsky()

        engine = sk.EngineDO(geometry=geometry, atmosphere=atmosphere, wavelengths=[440])
        rad1, wf1 = engine.calculate_radiance()
        engine.options['averagereferencepoint'] = True
        rad2, wf2 = engine.calculate_radiance()

        for r1, r2 in zip(rad1.flatten(), rad2.flatten()):
            self.assertAlmostEqual(r1, r2)

        for w1, w2 in zip(wf1.flatten(), wf2.flatten()):
            self.assertAlmostEqual(w1, w2)

    def test_lambertian(self):
        """ Continuity check to track changes in calculated radiances. 
        """
        geo = skif.GEODETIC_INSTANT()
        geo.FromSequence([0, 0, 0, 54372])

        clim_air = skif.ISKClimatology('MSIS90')
        opt_air = skif.ISKOpticalProperty('RAYLEIGH')

        engine = skif.ISKEngine("DO")
        engine.SetProperty('numstreams', 40)
        engine.SetProperty('numlayers', 200)

        engine.AddSpecies('SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3', clim_air, opt_air)
        engine.AddSpecies('SKCLIMATOLOGY_O3_CM3', skif.ISKClimatology("O3LABOW"),
                          skif.ISKOpticalProperty("O3_OSIRISRES"))

        engine.SetAtmosphericState(clim_air)

        sun = np.array([-0.190809, -0.98162718, 0])
        engine.SetProperty('sun', sun)

        obs = np.array([ -6880579.70829706, -39021706.60809562, 14407230.65919554])
        looks = np.array([[ 0.15042014, 0.87062868, -0.46837964],
                          [ 0.1518094, 0.87854755, -0.45288863],
                          [ 0.15315242, 0.8861988, -0.43725967],
                          [ 0.15444878, 0.89358011, -0.42149752],
                          [ 0.1556981, 0.90068923, -0.40560698],
                          [ 0.15689999, 0.90752398, -0.38959288],
                          [0.15805408, 0.9140823, -0.37346012],
                          [0.15916003, 0.92036218, -0.35721359],
                          [0.1602175, 0.9263617, -0.34085825],
                          [0.16122617, 0.93207905, -0.32439908]])
        mjd = 54372.5
        for los in np.reshape(looks, (-1, 3)):
            engine.AddLineOfSight(mjd, obs, los)

        engine.SetWavelengths([350, 400, 700])

        lambertian_brdf = skif.ISKBrdf('Lambertian')
        lambertian_brdf.SetProperty('Albedo', 0.3)

        engine.SetBRDF(lambertian_brdf)

        engine.InitializeModel()
        ok, rad = engine.CalculateRadiance()

        correct = np.array([
            [ 0.14010926, 0.13446559, 0.13185127, 0.12991796, 0.12810804, 0.12622069
            , 0.12416023, 0.12186823, 0.11928992, 0.11640004],
            [ 0.11911075, 0.11578273, 0.11526143, 0.11484618, 0.11409809, 0.11291455
            , 0.11126909, 0.10915426, 0.10656346, 0.10348698],
            [ 0.07810069, 0.08722341, 0.09215029, 0.0947986 , 0.09591001, 0.09584553
            , 0.09478171, 0.09282681, 0.08998259, 0.0862821 ]
        ])

        self.assertEqual(len(rad.flatten()), len(correct.flatten()))
        for r, c in zip(rad.flatten(), correct.flatten()):
            self.assertAlmostEqual(r, c)

    def test_pineforest(self):
        """ Continuity check to track changes in calculated radiances. 
        """
        geo = skif.GEODETIC_INSTANT()
        geo.FromSequence([0, 0, 0, 54372])

        clim_air = skif.ISKClimatology('MSIS90')
        opt_air = skif.ISKOpticalProperty('RAYLEIGH')

        engine = skif.ISKEngine("DO")
        engine.SetProperty('numstreams', 40)
        engine.SetProperty('numlayers', 200)

        engine.AddSpecies('SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3', clim_air, opt_air)
        engine.AddSpecies('SKCLIMATOLOGY_O3_CM3', skif.ISKClimatology("O3LABOW"),
                          skif.ISKOpticalProperty("O3_OSIRISRES"))

        engine.SetAtmosphericState(clim_air)

        sun = np.array([-0.190809, -0.98162718, 0])
        engine.SetProperty('sun', sun)

        obs = np.array([ -6880579.70829706, -39021706.60809562, 14407230.65919554])
        looks = np.array([[ 0.15042014, 0.87062868, -0.46837964],
                          [ 0.1518094, 0.87854755, -0.45288863],
                          [ 0.15315242, 0.8861988, -0.43725967],
                          [ 0.15444878, 0.89358011, -0.42149752],
                          [ 0.1556981, 0.90068923, -0.40560698],
                          [ 0.15689999, 0.90752398, -0.38959288],
                          [0.15805408, 0.9140823, -0.37346012],
                          [0.15916003, 0.92036218, -0.35721359],
                          [0.1602175, 0.9263617, -0.34085825],
                          [0.16122617, 0.93207905, -0.32439908]])
        mjd = 54372.5
        for los in np.reshape(looks, (-1, 3)):
            engine.AddLineOfSight(mjd, obs, los)

        engine.SetWavelengths([350, 400, 700])

        roujean_brdf = skif.ISKBrdf('Roujean')
        roujean_brdf.SetProperty('SetPredefinedParameters', 8)

        engine.SetBRDF(roujean_brdf)

        engine.InitializeModel()
        ok, rad = engine.CalculateRadiance()

        """
        For sasktranif <= 4.2.2 the following was the correct radiances
        correct = np.array([
            [ 0.11198657, 0.09518884, 0.08603374, 0.08006655, 0.07595647, 0.07311574
            , 0.07122159, 0.07007446, 0.0695374 , 0.06952421],
            [ 0.08177645, 0.06563946, 0.05805092, 0.05345452, 0.05043561, 0.04842605
            , 0.04713995, 0.04641629, 0.04616028, 0.04632018],
            [ 0.02297224, 0.01893762, 0.01716691, 0.01611499, 0.01542312, 0.01495173
            , 0.01463013, 0.01442208, 0.01429883, 0.014251  ]
        ])
        """
        correct = np.array([
            [ 0.11148904, 0.09434729, 0.08527606, 0.07959315, 0.07581134, 0.07323381
            , 0.07146295, 0.07025823, 0.06952026, 0.06923899],
            [ 0.08048155, 0.06390763, 0.05662706, 0.05260728, 0.05018315, 0.04862828
            , 0.04755065, 0.04672907, 0.0461309 , 0.04582443],
            [ 0.01898272, 0.01491508, 0.01419799, 0.01444746, 0.01494251, 0.01532985
            , 0.01539239, 0.01500276, 0.0142439 , 0.01330962]
        ])
        self.assertEqual(len(rad.flatten()), len(correct.flatten()))
        for r, c in zip(rad.flatten(), correct.flatten()):
            self.assertAlmostEqual(r, c)

    def test_elevated_surfaces(self):
        """ Check that elevated surfaces are working. Note that this test requires pysasktran to be installed.
        
        This check that engine holds to within 1% of radiances calculated with the beer lambert law for a purely 
        absorbing atmosphere.
        """
        try:
            import sasktran as sk
        except:
            import logging
            msg = "Elevated surfaces will not be tested because the 'sasktran' library is not installed. To install it" \
                  " enter: 'pip install sasktran -f https://arg.usask.ca/wheels/"

            logging.getLogger().log(logging.WARNING, msg)
            return

        wavelengths = [300, 310, 320, 330, 340, 350]
        latitude = 0
        longitude = -100
        mjd = 57906

        # Setup a purely absorbing atmosphere (with Lambertian(1) surface)
        optical_properties = sk.O3OSIRISRes() # ozone is purely absorbing
        climatology = sk.Labow()

        atmosphere = sk.Atmosphere()
        atmosphere.atmospheric_state = sk.MSIS90()
        atmosphere['O3'] = sk.Species(optical_properties, climatology)
        atmosphere.brdf = sk.Lambertian(1.0)

        # Calculate the optical depth of the atmosphere
        altitudes = np.linspace(100e3, 0, 501)
        center_altitudes = (altitudes[:-1] + altitudes[1:]) / 2
        box_width = np.diff(altitudes)
        number_density = climatology.get_parameter('SKCLIMATOLOGY_O3_CM3', latitude, longitude, center_altitudes, mjd)
        optical_depth = np.ones((len(center_altitudes), len(wavelengths))) * np.nan

        for i, (alt, width, nd) in enumerate(zip(center_altitudes, box_width, number_density)):
            cs = optical_properties.calculate_cross_sections(
                atmosphere.atmospheric_state,
                latitude, longitude, alt,
                57906,
                wavelengths
            ).total

            optical_depth[i, :] = cs * nd * (np.abs(width) * 100)

        optical_depth = np.cumsum(optical_depth, axis=0)
        optical_depth = np.insert(optical_depth, 0, np.zeros_like(optical_depth[0,:]), axis=0)

        # Configure geometry
        geometry = sk.Geometry()

        terminator = sk.Geodetic()
        terminator.from_lat_lon_alt(latitude, longitude, 0)

        obs = sk.Geodetic()
        obs.from_lat_lon_alt(latitude, longitude, 150e3)

        look = (terminator.location - obs.location)
        look = look / np.linalg.norm(look)

        geometry.lines_of_sight = [
            sk.LineOfSight(look_vector=look, observer=obs.location, mjd=mjd)
        ]

        sun = sk.Geodetic()
        sun.from_lat_lon_alt(latitude, longitude, 200e3)
        geometry.sun = sun.local_up

        for surface_height in [0, 5e3, 10e3, 15e3, 20e3, 25e3, 30e3, 35e3]:
            engine = sk.EngineDO(atmosphere=atmosphere, geometry=geometry, wavelengths=wavelengths)
            step = np.min(np.abs(box_width))
            engine.options['altitudegrid'] = np.arange(surface_height, 100e3 + step - 1, step)

            idx = [idx for idx, alt in enumerate(altitudes) if alt <= surface_height][0]
            correct = np.exp(-2.0 * optical_depth[idx,:]) / np.pi
            calculated = engine.calculate_radiance().flatten()

            diff = np.abs(correct - calculated) / (0.5 * (correct + calculated))
            for wav, d in zip(wavelengths, diff):
                self.assertTrue(d < 1e-2, "Wavelength {} nm failed at a surface height of {} m with a {:.2f}% difference".format(
                    wav, surface_height, 100.0*d
                ))

if __name__ == "__main__":
    tests = QuickTest()
    unittest.main()